@@ -4,21 +4,41 @@
 # LICENSE file in the root directory of this source tree.
 
 import asyncio
-import inspect
 import functools
-from typing import List, Callable, Tuple, Any, Dict
+import inspect
+from inspect import Traceback
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+    Type,
+    Union,
+)
+from unittest.mock import Mock
+
 import testslide
+from testslide.lib import _validate_return_type, _wrap_signature_and_type_validation
 from testslide.strict_mock import StrictMock
-from testslide.lib import _wrap_signature_and_type_validation, _validate_return_type
-from .patch import _patch, _is_instance_method
-from .lib import _bail_if_private
+
+from .lib import CoroutineValueError, _bail_if_private
+from .patch import _is_instance_method, _patch
+
+if TYPE_CHECKING:
+    from testslide.matchers import RegexMatches  # noqa: F401
+    from testslide.mock_constructor import _MockConstructorDSL  # noqa: F401
 
 
-def mock_callable(target, method, allow_private=False, type_validation=True):
-    caller_frame = inspect.currentframe().f_back
+def mock_callable(
+    target: Any, method: str, allow_private: bool = False, type_validation: bool = True
+) -> "_MockCallableDSL":
+    caller_frame = inspect.currentframe().f_back  # type: ignore
     # loading the context ends up reading files from disk and that might block
     # the event loop, so we don't do it.
-    caller_frame_info = inspect.getframeinfo(caller_frame, context=0)
+    caller_frame_info = inspect.getframeinfo(caller_frame, context=0)  # type: ignore
     return _MockCallableDSL(
         target,
         method,
@@ -29,16 +49,16 @@ def mock_callable(target, method, allow_private=False, type_validation=True):
 
 
 def mock_async_callable(
-    target,
-    method,
-    callable_returns_coroutine=False,
-    allow_private=False,
-    type_validation=True,
-):
-    caller_frame = inspect.currentframe().f_back
+    target: Union[type, str],
+    method: str,
+    callable_returns_coroutine: bool = False,
+    allow_private: bool = False,
+    type_validation: bool = True,
+) -> "_MockAsyncCallableDSL":
+    caller_frame = inspect.currentframe().f_back  # type: ignore
     # loading the context ends up reading files from disk and that might block
     # the event loop, so we don't do it.
-    caller_frame_info = inspect.getframeinfo(caller_frame, context=0)
+    caller_frame_info = inspect.getframeinfo(caller_frame, context=0)  # type: ignore
     return _MockAsyncCallableDSL(
         target,
         method,
@@ -49,10 +69,10 @@ def mock_async_callable(
     )
 
 
-_unpatchers = []  # type: List[Callable]  # noqa T484
+_unpatchers: List[Callable] = []  # noqa T484
 
 
-def _default_register_assertion(assertion):
+def _default_register_assertion(assertion: Callable) -> None:
     """
     This method must be redefined by the test framework using mock_callable().
     It will be called when a new assertion is defined, passing a callable as an
@@ -63,12 +83,12 @@ def _default_register_assertion(assertion):
 
 
 register_assertion = _default_register_assertion
-_call_order_assertion_registered = False
+_call_order_assertion_registered: bool = False
 _received_ordered_calls: List[Tuple[Any, str, "_BaseRunner"]] = []
 _expected_ordered_calls: List[Tuple[Any, str, "_BaseRunner"]] = []
 
 
-def unpatch_all_callable_mocks():
+def unpatch_all_callable_mocks() -> None:
     """
     This method must be called after every test unconditionally to remove all
     active mock_callable() patches.
@@ -93,12 +113,12 @@ def unpatch_all_callable_mocks():
         )
 
 
-def _is_setup():
+def _is_setup() -> bool:
     global register_assertion, _default_register_assertion
     return register_assertion is not _default_register_assertion
 
 
-def _format_target(target):
+def _format_target(target: Union[str, type]) -> str:
     if hasattr(target, "__repr__"):
         return repr(target)
     else:
@@ -107,7 +127,7 @@ def _format_target(target):
         )
 
 
-def _format_args(indent, *args, **kwargs):
+def _format_args(indent: int, *args: Any, **kwargs: Any) -> str:
     indentation = "  " * indent
     s = ""
     if args:
@@ -123,9 +143,9 @@ def _format_args(indent, *args, **kwargs):
     return s
 
 
-def _is_coroutine(obj):
+def _is_coroutine(obj: Any) -> bool:
 
-    return inspect.iscoroutine(obj) or isinstance(obj, asyncio.coroutines.CoroWrapper)
+    return inspect.iscoroutine(obj) or isinstance(obj, asyncio.coroutines.CoroWrapper)  # type: ignore
 
 
 ##
@@ -173,17 +193,19 @@ class NotACoroutine(BaseException):
 class _BaseRunner:
     TYPE_VALIDATION = True
 
-    def __init__(self, target, method, original_callable):
+    def __init__(
+        self, target: Any, method: str, original_callable: Union[Callable, Mock]
+    ) -> None:
         self.target = target
         self.method = method
         self.original_callable = original_callable
-        self.accepted_args = None
+        self.accepted_args: Optional[Tuple[Any, Any]] = None
 
-        self._call_count = 0
-        self._max_calls = None
+        self._call_count: int = 0
+        self._max_calls: Optional[int] = None
         self._has_order_assertion = False
 
-    def register_call(self, *args, **kwargs):
+    def register_call(self, *args: Any, **kwargs: Any) -> None:
         global _received_ordered_calls
 
         if self._has_order_assertion:
@@ -192,18 +214,18 @@ class _BaseRunner:
         self.inc_call_count()
 
     @property
-    def call_count(self):
+    def call_count(self) -> int:
         return self._call_count
 
     @property
-    def max_calls(self):
+    def max_calls(self) -> Optional[int]:
         return self._max_calls
 
-    def _set_max_calls(self, times):
+    def _set_max_calls(self, times: int) -> None:
         if not self._max_calls or times < self._max_calls:
             self._max_calls = times
 
-    def inc_call_count(self):
+    def inc_call_count(self) -> None:
         self._call_count += 1
         if self.max_calls and self._call_count > self.max_calls:
             raise UnexpectedCallReceived(
@@ -220,11 +242,11 @@ class _BaseRunner:
                 )
             )
 
-    def add_accepted_args(self, *args, **kwargs):
+    def add_accepted_args(self, *args: Any, **kwargs: Any) -> None:
         # TODO validate if args match callable signature
         self.accepted_args = (args, kwargs)
 
-    def can_accept_args(self, *args, **kwargs):
+    def can_accept_args(self, *args: Any, **kwargs: Any) -> bool:
         if self.accepted_args:
             if self.accepted_args == (args, kwargs):
                 return True
@@ -232,7 +254,7 @@ class _BaseRunner:
         else:
             return True
 
-    def _args_message(self):
+    def _args_message(self) -> str:
         if self.accepted_args:
             return "arguments:\n{}".format(
                 _format_args(2, *self.accepted_args[0], **self.accepted_args[1])
@@ -240,10 +262,10 @@ class _BaseRunner:
         else:
             return "any arguments "
 
-    def add_exact_calls_assertion(self, times):
+    def add_exact_calls_assertion(self, times: int) -> None:
         self._set_max_calls(times)
 
-        def assertion():
+        def assertion() -> None:
             if times != self.call_count:
                 raise AssertionError(
                     (
@@ -262,8 +284,8 @@ class _BaseRunner:
 
         register_assertion(assertion)
 
-    def add_at_least_calls_assertion(self, times):
-        def assertion():
+    def add_at_least_calls_assertion(self, times: int) -> None:
+        def assertion() -> None:
             if self.call_count < times:
                 raise AssertionError(
                     (
@@ -282,10 +304,10 @@ class _BaseRunner:
 
         register_assertion(assertion)
 
-    def add_at_most_calls_assertion(self, times):
+    def add_at_most_calls_assertion(self, times: int) -> None:
         self._set_max_calls(times)
 
-        def assertion():
+        def assertion() -> None:
             if not self.call_count or self.call_count > times:
                 raise AssertionError(
                     (
@@ -304,12 +326,12 @@ class _BaseRunner:
 
         register_assertion(assertion)
 
-    def add_call_order_assertion(self):
+    def add_call_order_assertion(self) -> None:
         global _call_order_assertion_registered, _received_ordered_calls, _expected_ordered_calls
 
         if not _call_order_assertion_registered:
 
-            def assertion():
+            def assertion() -> None:
                 if _received_ordered_calls != _expected_ordered_calls:
                     raise AssertionError(
                         (
@@ -354,32 +376,50 @@ class _BaseRunner:
 
 
 class _Runner(_BaseRunner):
-    def run(self, *args, **kwargs):
+    def run(self, *args: Any, **kwargs: Any) -> None:
         super().register_call(*args, **kwargs)
 
 
 class _AsyncRunner(_BaseRunner):
-    async def run(self, *args, **kwargs):
+    async def run(self, *args: Any, **kwargs: Any) -> None:
         super().register_call(*args, **kwargs)
 
 
 class _ReturnValueRunner(_Runner):
-    def __init__(self, target, method, original_callable, value):
+    def __init__(
+        self,
+        target: Any,
+        method: str,
+        original_callable: Union[Callable, Mock],
+        value: Optional[Any],
+        allow_coro: bool = False,
+    ) -> None:
         super().__init__(target, method, original_callable)
+        if not allow_coro and _is_coroutine(value):
+            raise CoroutineValueError()
         self.return_value = value
 
-    def run(self, *args, **kwargs):
+    def run(self, *args: Any, **kwargs: Any) -> Optional[Any]:
         super().run(*args, **kwargs)
         return self.return_value
 
 
 class _ReturnValuesRunner(_Runner):
-    def __init__(self, target, method, original_callable, values_list):
+    def __init__(
+        self,
+        target: Union[type, str],
+        method: str,
+        original_callable: Union[Callable[..., Any], Mock],
+        values_list: List[Any],
+        allow_coro: bool = False,
+    ) -> None:
         super(_ReturnValuesRunner, self).__init__(target, method, original_callable)
         # Reverse original list for popping efficiency
+        if not allow_coro and any(_is_coroutine(rv) for rv in values_list):
+            raise CoroutineValueError()
         self.values_list = list(reversed(values_list))
 
-    def run(self, *args, **kwargs):
+    def run(self, *args: Any, **kwargs: Any) -> Any:
         super(_ReturnValuesRunner, self).run(*args, **kwargs)
         if self.values_list:
             return self.values_list.pop()
@@ -390,15 +430,24 @@ class _ReturnValuesRunner(_Runner):
 class _YieldValuesRunner(_Runner):
     TYPE_VALIDATION = False
 
-    def __init__(self, target, method, original_callable, values_list):
+    def __init__(
+        self,
+        target: Union[type, str],
+        method: str,
+        original_callable: Union[Callable[..., Any], Mock],
+        values_list: List[Any],
+        allow_coro: bool = False,
+    ) -> None:
         super(_YieldValuesRunner, self).__init__(target, method, original_callable)
         self.values_list = values_list
         self.index = 0
+        if not allow_coro and any(_is_coroutine(rv) for rv in values_list):
+            raise CoroutineValueError()
 
-    def __iter__(self):
+    def __iter__(self) -> "_YieldValuesRunner":
         return self
 
-    def __next__(self):
+    def __next__(self) -> Any:
         try:
             item = self.values_list[self.index]
         except IndexError:
@@ -406,37 +455,60 @@ class _YieldValuesRunner(_Runner):
         self.index += 1
         return item
 
-    def run(self, *args, **kwargs):
+    def run(self, *args: Any, **kwargs: Any) -> "_YieldValuesRunner":  # type: ignore
         super(_YieldValuesRunner, self).run(*args, **kwargs)
         return self
 
 
 class _RaiseRunner(_Runner):
-    def __init__(self, target, method, original_callable, exception):
+    def __init__(
+        self,
+        target: Union[type, str],
+        method: str,
+        original_callable: Union[Callable[..., Any], Mock],
+        exception: Union[Exception, BaseException],
+    ) -> None:
         super(_RaiseRunner, self).__init__(target, method, original_callable)
         self.exception = exception
 
-    def run(self, *args, **kwargs):
+    def run(self, *args: Any, **kwargs: Any) -> None:
         super(_RaiseRunner, self).run(*args, **kwargs)
         raise self.exception
 
 
 class _ImplementationRunner(_Runner):
-    def __init__(self, target, method, original_callable, new_implementation):
+    def __init__(
+        self,
+        target: Union[type, str],
+        method: str,
+        original_callable: Union[Callable[..., Any], Mock],
+        new_implementation: Callable,
+        allow_coro: bool = False,
+    ) -> None:
         super(_ImplementationRunner, self).__init__(target, method, original_callable)
         self.new_implementation = new_implementation
+        self._allow_coro = allow_coro
 
-    def run(self, *args, **kwargs):
+    def run(self, *args: Any, **kwargs: Any) -> Optional[Any]:
         super(_ImplementationRunner, self).run(*args, **kwargs)
-        return self.new_implementation(*args, **kwargs)
+        new_impl = self.new_implementation(*args, **kwargs)
+        if not self._allow_coro and _is_coroutine(new_impl):
+            raise CoroutineValueError()
+        return new_impl
 
 
 class _AsyncImplementationRunner(_AsyncRunner):
-    def __init__(self, target, method, original_callable, new_implementation):
+    def __init__(
+        self,
+        target: Union[type, str],
+        method: str,
+        original_callable: Union[Callable[..., Any], Mock],
+        new_implementation: Callable,
+    ) -> None:
         super().__init__(target, method, original_callable)
         self.new_implementation = new_implementation
 
-    async def run(self, *args, **kwargs):
+    async def run(self, *args: Any, **kwargs: Any) -> Optional[Any]:
         await super().run(*args, **kwargs)
         coro = self.new_implementation(*args, **kwargs)
         if not _is_coroutine(coro):
@@ -448,13 +520,13 @@ class _AsyncImplementationRunner(_AsyncRunner):
 
 
 class _CallOriginalRunner(_Runner):
-    def run(self, *args, **kwargs):
+    def run(self, *args: Any, **kwargs: Any) -> Optional[Any]:
         super(_CallOriginalRunner, self).run(*args, **kwargs)
         return self.original_callable(*args, **kwargs)
 
 
 class _AsyncCallOriginalRunner(_AsyncRunner):
-    async def run(self, *args, **kwargs):
+    async def run(self, *args: Any, **kwargs: Any) -> Optional[Any]:
         await super().run(*args, **kwargs)
         return await self.original_callable(*args, **kwargs)
 
@@ -466,22 +538,27 @@ class _AsyncCallOriginalRunner(_AsyncRunner):
 
 class _CallableMock(object):
     def __init__(
-        self, target, method, caller_frame_info, is_async=False, type_validation=True,
-    ):
+        self,
+        target: Any,
+        method: str,
+        caller_frame_info: Traceback,
+        is_async: bool = False,
+        type_validation: bool = True,
+    ) -> None:
         self.target = target
         self.method = method
-        self.runners = []
+        self.runners: List[_BaseRunner] = []
         self.is_async = is_async
         self.type_validation = type_validation
         self.caller_frame_info = caller_frame_info
 
-    def _get_runner(self, *args, **kwargs):
+    def _get_runner(self, *args: Any, **kwargs: Any) -> Any:
         for runner in self.runners:
             if runner.can_accept_args(*args, **kwargs):
                 return runner
         return None
 
-    def _validate_return_type(self, runner, value):
+    def _validate_return_type(self, runner: _BaseRunner, value: Any) -> None:
         if (
             self.type_validation
             and runner.TYPE_VALIDATION
@@ -491,13 +568,13 @@ class _CallableMock(object):
                 runner.original_callable, value, self.caller_frame_info
             )
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: Any, **kwargs: Any) -> Optional[Any]:
         runner = self._get_runner(*args, **kwargs)
         if runner:
             if self.is_async:
                 if isinstance(runner, _AsyncRunner):
 
-                    async def async_wrapper(*args, **kwargs):
+                    async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                         value = await runner.run(*args, **kwargs)
                         self._validate_return_type(runner, value)
                         return value
@@ -505,7 +582,7 @@ class _CallableMock(object):
                     value = async_wrapper(*args, **kwargs)
                 else:
 
-                    async def async_wrapper(*args, **kwargs):
+                    async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                         value = runner.run(*args, **kwargs)
                         self._validate_return_type(runner, value)
                         return value
@@ -537,7 +614,7 @@ class _CallableMock(object):
             raise UndefinedBehaviorForCall(ex_msg)
 
     @property
-    def _registered_calls(self):
+    def _registered_calls(self) -> Any:
         return [runner.accepted_args for runner in self.runners if runner.accepted_args]
 
 
@@ -548,15 +625,18 @@ class _CallableMock(object):
 
 class _MockCallableDSL(object):
 
-    CALLABLE_MOCKS: Dict[int, _CallableMock] = {}
+    CALLABLE_MOCKS: Dict[
+        Union[int, Tuple[int, str]], Union[Callable[[Type[object]], Any]]
+    ] = {}
+    _NAME: str = "mock_callable"
 
     def _validate_patch(
         self,
-        name="mock_callable",
-        other_name="mock_async_callable",
-        coroutine_function=False,
-        callable_returns_coroutine=False,
-    ):
+        name: str = "mock_callable",
+        other_name: str = "mock_async_callable",
+        coroutine_function: bool = False,
+        callable_returns_coroutine: bool = False,
+    ) -> None:
         if self._method == "__new__":
             raise ValueError(
                 f"Mocking __new__ is not allowed with {name}(), please use "
@@ -639,7 +719,9 @@ class _MockCallableDSL(object):
                     f"use {other_name}() instead."
                 )
 
-    def _patch(self, new_value):
+    def _patch(
+        self, new_value: Union[Callable, _CallableMock]
+    ) -> Union[Tuple[Callable, Callable], Tuple[Mock, Callable], Tuple[None, Callable]]:
         self._validate_patch()
 
         if isinstance(self._target, StrictMock):
@@ -656,7 +738,7 @@ class _MockCallableDSL(object):
         restore_value = self._target.__dict__.get(self._method, None)
 
         if inspect.isclass(self._target):
-            new_value = staticmethod(new_value)
+            new_value = staticmethod(new_value)  # type: ignore
 
         unpatcher = _patch(
             self._target, self._method, new_value, restore, restore_value
@@ -664,7 +746,7 @@ class _MockCallableDSL(object):
 
         return original_callable, unpatcher
 
-    def _get_callable_mock(self):
+    def _get_callable_mock(self) -> _CallableMock:
         return _CallableMock(
             self._original_target,
             self._method,
@@ -674,27 +756,33 @@ class _MockCallableDSL(object):
 
     def __init__(
         self,
-        target,
-        method,
-        caller_frame_info,
-        callable_mock=None,
-        original_callable=None,
-        allow_private=False,
-        type_validation=True,
-    ):
+        target: Any,
+        method: str,
+        caller_frame_info: Traceback,
+        callable_mock: Union[Callable[[Type[object]], Any], _CallableMock, None] = None,
+        original_callable: Optional[Callable] = None,
+        allow_private: bool = False,
+        type_validation: bool = True,
+    ) -> None:
         if not _is_setup():
             raise RuntimeError(
-                "mock_callable() was not setup correctly before usage. "
-                "Please refer to the documentation at http://testslide.readthedocs.io/ "
-                "to learn how to properly do that."
+                "TestSlide was not correctly setup before usage!\n"
+                "This error happens when mock_callable, mock_async_callable or "
+                "mock_constructor are attempted to be used without correct "
+                "test framework integration, meaning unpatching and "
+                "assertions will not work as expected.\n"
+                "A common scenario for this is when testslide.TestCase is "
+                "subclassed with setUp() overridden but super().setUp() was not "
+                "called."
             )
         self._original_target = target
         self._method = method
-        self._runner = None
-        self._next_runner_accepted_args = None
+        self._runner: Optional[_BaseRunner] = None
+        self._next_runner_accepted_args: Any = None
         self.allow_private = allow_private
         self.type_validation = type_validation
         self.caller_frame_info = caller_frame_info
+        self._allow_coro = False
         if isinstance(target, str):
             self._target = testslide._importer(target)
         else:
@@ -711,7 +799,7 @@ class _MockCallableDSL(object):
             self.CALLABLE_MOCKS[target_method_id] = callable_mock
             self._callable_mock = callable_mock
 
-            def del_callable_mock():
+            def del_callable_mock() -> None:
                 del self.CALLABLE_MOCKS[target_method_id]
 
             _unpatchers.append(del_callable_mock)
@@ -720,12 +808,12 @@ class _MockCallableDSL(object):
                 original_callable, unpatcher = self._patch(callable_mock)
                 _unpatchers.append(unpatcher)
             self._original_callable = original_callable
-            callable_mock.original_callable = original_callable
+            callable_mock.original_callable = original_callable  # type: ignore
         else:
             self._callable_mock = self.CALLABLE_MOCKS[target_method_id]
-            self._original_callable = self._callable_mock.original_callable
+            self._original_callable = self._callable_mock.original_callable  # type: ignore
 
-    def _add_runner(self, runner):
+    def _add_runner(self, runner: _BaseRunner) -> None:
         if self._runner:
             raise ValueError(
                 "Can't define more than one behavior using the same "
@@ -737,9 +825,9 @@ class _MockCallableDSL(object):
             self._next_runner_accepted_args = None
             runner.add_accepted_args(*args, **kwargs)
         self._runner = runner
-        self._callable_mock.runners.insert(0, runner)
+        self._callable_mock.runners.insert(0, runner)  # type: ignore
 
-    def _assert_runner(self):
+    def _assert_runner(self) -> None:
         if not self._runner:
             raise ValueError(
                 "You must first define a behavior. Eg: "
@@ -747,12 +835,23 @@ class _MockCallableDSL(object):
                 ".to_return_value(value)"
                 ".and_assert_called_exactly(times)"
             )
+        if self._runner._call_count > 0:
+            raise ValueError(
+                f"No extra configuration is allowed after {self._NAME} "
+                f"receives its first call, it received {self._runner._call_count} "
+                f"call{'s' if self._runner._call_count > 1 else ''} already. "
+                "You should instead define it all at once, "
+                f"eg: self.{self._NAME}(target, 'func')"
+                ".to_return_value(value).and_assert_called_once()"
+            )
 
     ##
     ## Arguments
     ##
 
-    def for_call(self, *args, **kwargs):
+    def for_call(
+        self, *args: Any, **kwargs: Any
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Filter for only calls like this.
         """
@@ -766,18 +865,26 @@ class _MockCallableDSL(object):
     ## Behavior
     ##
 
-    def to_return_value(self, value):
+    def to_return_value(
+        self, value: Optional[Union[bool, str, int]]
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Always return given value.
         """
         self._add_runner(
             _ReturnValueRunner(
-                self._original_target, self._method, self._original_callable, value
+                self._original_target,
+                self._method,
+                self._original_callable,  # type: ignore
+                value,
+                self._allow_coro,
             )
         )
         return self
 
-    def to_return_values(self, values_list):
+    def to_return_values(
+        self, values_list: Union[List[str], List[Union[int, str]]]
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         For each call, return each value from given list in order.
         When list is exhausted, goes to the next behavior set.
@@ -788,13 +895,16 @@ class _MockCallableDSL(object):
             _ReturnValuesRunner(
                 self._original_target,
                 self._method,
-                self._original_callable,
+                self._original_callable,  # type: ignore
                 values_list,
+                self._allow_coro,
             )
         )
         return self
 
-    def to_yield_values(self, values_list):
+    def to_yield_values(
+        self, values_list: List[str]
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Callable will return an iterator what will yield each value from the
         given list.
@@ -805,26 +915,29 @@ class _MockCallableDSL(object):
             _YieldValuesRunner(
                 self._original_target,
                 self._method,
-                self._original_callable,
+                self._original_callable,  # type: ignore
                 values_list,
+                self._allow_coro,
             )
         )
         return self
 
-    def to_raise(self, ex):
+    def to_raise(
+        self, ex: Union[UnexpectedCallReceived, Type[RuntimeError], RuntimeError]
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Raises given exception class or exception instance.
         """
         if isinstance(ex, BaseException):
             self._add_runner(
                 _RaiseRunner(
-                    self._original_target, self._method, self._original_callable, ex
+                    self._original_target, self._method, self._original_callable, ex  # type: ignore
                 )
             )
         elif isinstance(ex(), BaseException):
             self._add_runner(
                 _RaiseRunner(
-                    self._original_target, self._method, self._original_callable, ex()
+                    self._original_target, self._method, self._original_callable, ex()  # type: ignore
                 )
             )
         else:
@@ -833,7 +946,9 @@ class _MockCallableDSL(object):
             )
         return self
 
-    def with_implementation(self, func):
+    def with_implementation(
+        self, func: Callable
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Replace callable by given function.
         """
@@ -841,12 +956,18 @@ class _MockCallableDSL(object):
             raise ValueError("{} must be callable.".format(func))
         self._add_runner(
             _ImplementationRunner(
-                self._original_target, self._method, self._original_callable, func
+                self._original_target,
+                self._method,
+                self._original_callable,  # type: ignore
+                func,
+                self._allow_coro,
             )
         )
         return self
 
-    def with_wrapper(self, func):
+    def with_wrapper(
+        self, func: Callable
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Replace callable with given wrapper function, that will be called as:
 
@@ -862,17 +983,23 @@ class _MockCallableDSL(object):
             raise ValueError("Can not wrap original callable that does not exist.")
 
         @functools.wraps(func)
-        def wrapper(*args, **kwargs):
+        def wrapper(*args: Any, **kwargs: Any) -> Any:
             return func(self._original_callable, *args, **kwargs)
 
         self._add_runner(
             _ImplementationRunner(
-                self._original_target, self._method, self._original_callable, wrapper
+                self._original_target,
+                self._method,
+                self._original_callable,
+                wrapper,
+                self._allow_coro,
             )
         )
         return self
 
-    def to_call_original(self):
+    def to_call_original(
+        self,
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Calls the original callable implementation, instead of mocking it. This is
         useful for example, if you want to by default call the original implementation,
@@ -891,7 +1018,9 @@ class _MockCallableDSL(object):
     ## Call Assertions
     ##
 
-    def and_assert_called_exactly(self, count):
+    def and_assert_called_exactly(
+        self, count: int
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Assert that there were exactly the given number of calls.
 
@@ -909,73 +1038,90 @@ class _MockCallableDSL(object):
                         )
                     )
                 )
-        self._runner.add_exact_calls_assertion(count)
+        self._runner.add_exact_calls_assertion(count)  # type: ignore
         return self
 
-    def and_assert_not_called(self):
+    def and_assert_not_called(
+        self,
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Short for and_assert_called_exactly(0)
         """
         return self.and_assert_called_exactly(0)
 
-    def and_assert_called_once(self):
+    def and_assert_called_once(
+        self,
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Short for and_assert_called_exactly(1)
         """
         return self.and_assert_called_exactly(1)
 
-    def and_assert_called_twice(self):
+    def and_assert_called_twice(
+        self,
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Short for and_assert_called_exactly(2)
         """
         return self.and_assert_called_exactly(2)
 
-    def and_assert_called_at_least(self, count):
+    def and_assert_called_at_least(
+        self, count: int
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Assert that there at least the given number of calls.
         """
         if count < 1:
             raise ValueError("times must be >= 1")
         self._assert_runner()
-        self._runner.add_at_least_calls_assertion(count)
+        self._runner.add_at_least_calls_assertion(count)  # type: ignore
         return self
 
-    def and_assert_called_at_most(self, count):
+    def and_assert_called_at_most(
+        self, count: int
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Assert that there at most the given number of calls.
         """
         if count < 1:
             raise ValueError("times must be >= 1")
         self._assert_runner()
-        self._runner.add_at_most_calls_assertion(count)
+        self._runner.add_at_most_calls_assertion(count)  # type: ignore
         return self
 
-    def and_assert_called(self):
+    def and_assert_called(
+        self,
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Short for self.and_assert_called_at_least(1).
         """
         return self.and_assert_called_at_least(1)
 
-    def and_assert_called_ordered(self):
+    def and_assert_called_ordered(
+        self,
+    ) -> Union["_MockCallableDSL", "_MockAsyncCallableDSL", "_MockConstructorDSL"]:
         """
         Assert that multiple calls, potentially to different mock_callable()
         targets, happened in the order defined.
         """
         self._assert_runner()
-        self._runner.add_call_order_assertion()
+        self._runner.add_call_order_assertion()  # type: ignore
         return self
 
 
 class _MockAsyncCallableDSL(_MockCallableDSL):
+
+    _NAME: str = "mock_async_callable"
+
     def __init__(
         self,
-        target,
-        method,
-        caller_frame_info,
-        callable_returns_coroutine,
-        allow_private=False,
-        type_validation=True,
-    ):
+        target: Union[str, type],
+        method: str,
+        caller_frame_info: Traceback,
+        callable_returns_coroutine: bool,
+        allow_private: bool = False,
+        type_validation: bool = True,
+    ) -> None:
         self._callable_returns_coroutine = callable_returns_coroutine
         super().__init__(
             target,
@@ -984,16 +1130,17 @@ class _MockAsyncCallableDSL(_MockCallableDSL):
             allow_private=allow_private,
             type_validation=type_validation,
         )
+        self._allow_coro = True
 
-    def _validate_patch(self):
+    def _validate_patch(self) -> None:  # type: ignore
         return super()._validate_patch(
-            name="mock_async_callable",
+            name=self._NAME,
             other_name="mock_callable",
             coroutine_function=True,
             callable_returns_coroutine=self._callable_returns_coroutine,
         )
 
-    def _get_callable_mock(self):
+    def _get_callable_mock(self) -> _CallableMock:
         return _CallableMock(
             self._original_target,
             self._method,
@@ -1002,7 +1149,7 @@ class _MockAsyncCallableDSL(_MockCallableDSL):
             type_validation=self.type_validation,
         )
 
-    def with_implementation(self, func):
+    def with_implementation(self, func: Callable) -> "_MockAsyncCallableDSL":
         """
         Replace callable by given async function.
         """
@@ -1010,12 +1157,12 @@ class _MockAsyncCallableDSL(_MockCallableDSL):
             raise ValueError("{} must be callable.".format(func))
         self._add_runner(
             _AsyncImplementationRunner(
-                self._original_target, self._method, self._original_callable, func
+                self._original_target, self._method, self._original_callable, func  # type: ignore
             )
         )
         return self
 
-    def with_wrapper(self, func):
+    def with_wrapper(self, func: Callable) -> "_MockAsyncCallableDSL":
         """
         Replace callable with given wrapper async function, that will be called as:
 
@@ -1031,7 +1178,7 @@ class _MockAsyncCallableDSL(_MockCallableDSL):
             raise ValueError("Can not wrap original callable that does not exist.")
 
         @functools.wraps(func)
-        async def wrapper(*args, **kwargs):
+        async def wrapper(*args: Any, **kwargs: Any) -> Any:
             coro = func(self._original_callable, *args, **kwargs)
             if not _is_coroutine(coro):
                 raise NotACoroutine(
@@ -1047,7 +1194,7 @@ class _MockAsyncCallableDSL(_MockCallableDSL):
         )
         return self
 
-    def to_call_original(self):
+    def to_call_original(self) -> "_MockAsyncCallableDSL":
         """
         Calls the original callable implementation, instead of mocking it. This is
         useful for example, if you want to by default call the original implementation,
