@@ -6,15 +6,20 @@
 import inspect
 import io
 import os
-import psutil
 import random
 import sys
 import time
 import traceback
+from contextlib import redirect_stderr, redirect_stdout
+from importlib import import_module
+from typing import Any, Callable, Dict, List, Optional, Pattern, Union, cast
 
-from . import AggregatedExceptions, Skip, _ExampleRunner
-from contextlib import redirect_stdout, redirect_stderr
+import psutil
+import pygments
+import pygments.formatters
+import pygments.lexers
 
+from . import AggregatedExceptions, Context, Example, Skip, _ExampleRunner
 
 ##
 ## Base
@@ -29,39 +34,47 @@ class BaseFormatter:
 
     def __init__(
         self,
-        force_color=False,
-        import_secs=None,
-        trim_path_prefix=None,
-        show_testslide_stack_trace=False,
-        dsl_debug=False,
-    ):
+        import_module_names: List[str],
+        force_color: bool = False,
+        import_secs: Optional[float] = None,
+        trim_path_prefix: Optional[str] = None,
+        show_testslide_stack_trace: bool = False,
+        dsl_debug: bool = False,
+    ) -> None:
+        self.import_module_names = import_module_names
         self.force_color = force_color
         self.import_secs = import_secs
         self._import_secs_warn = True
         self.trim_path_prefix = trim_path_prefix
         self.show_testslide_stack_trace = show_testslide_stack_trace
         self.dsl_debug = dsl_debug
-        self.current_hierarchy = []
-        self.results = {"success": [], "fail": [], "skip": []}
+        self.current_hierarchy: List[Context] = []
+        self.results: Dict[
+            str, List[Union[Example, Dict[str, Union[Example, BaseException]]]]
+        ] = {
+            "success": [],
+            "fail": [],
+            "skip": [],
+        }
         self.start_time = psutil.Process(os.getpid()).create_time()
-        self.end_time = None
-        self.duration_secs = None
+        self.end_time: Optional[float] = None
+        self.duration_secs: Optional[float] = None
 
     # Example Discovery
 
-    def discovery_start(self):
+    def discovery_start(self) -> None:
         """
         To be called before example discovery starts.
         """
         pass
 
-    def example_discovered(self, example):
+    def example_discovered(self, example: Example) -> None:
         """
         To be called when a new example is discovered.
         """
         print(example.full_name)
 
-    def discovery_finish(self):
+    def discovery_finish(self) -> None:
         """
         To be called before example discovery finishes.
         """
@@ -69,7 +82,7 @@ class BaseFormatter:
 
     # Test Execution
 
-    def start(self, example):
+    def start(self, example: Example) -> None:
         """
         To be called before each example execution.
         """
@@ -83,38 +96,38 @@ class BaseFormatter:
         self.new_example(example)
         self.current_hierarchy = example.context.hierarchy
 
-    def new_context(self, context):
+    def new_context(self, context: Context) -> None:
         """
         Called before an example execution, when its context is different from
         previous executed example.
         """
         pass
 
-    def new_example(self, example):
+    def new_example(self, example: Example) -> None:
         """
         Called before an example execution.
         """
         pass
 
-    def success(self, example):
+    def success(self, example: Example) -> None:
         """
         Called when an example was Successfuly executed.
         """
         self.results["success"].append(example)
 
-    def fail(self, example, exception):
+    def fail(self, example: Example, exception: BaseException) -> None:
         """
         Called when an example failed on execution.
         """
         self.results["fail"].append({"example": example, "exception": exception})
 
-    def skip(self, example):
+    def skip(self, example: Example) -> None:
         """
         Called when an example had the execution skipped.
         """
         self.results["skip"].append(example)
 
-    def finish(self, not_executed_examples):
+    def finish(self, not_executed_examples: List[Example]) -> None:
         """
         Called when all examples finished execution.
         """
@@ -123,25 +136,25 @@ class BaseFormatter:
 
     # DSL
 
-    def dsl_example(self, example, code):
+    def dsl_example(self, example: Example, code: Callable) -> None:
         pass
 
-    def dsl_before(self, example, code):
+    def dsl_before(self, example: Example, code: Callable) -> None:
         pass
 
-    def dsl_after(self, example, code):
+    def dsl_after(self, example: Example, code: Callable) -> None:
         pass
 
-    def dsl_around(self, example, code):
+    def dsl_around(self, example: Example, code: Callable) -> None:
         pass
 
-    def dsl_memoize(self, example, code):
+    def dsl_memoize(self, example: Example, code: Callable) -> None:
         pass
 
-    def dsl_memoize_before(self, example, code):
+    def dsl_memoize_before(self, example: Example, code: Callable) -> None:
         pass
 
-    def dsl_function(self, example, code):
+    def dsl_function(self, example: Example, code: Callable) -> None:
         pass
 
 
@@ -151,9 +164,15 @@ class BaseFormatter:
 
 
 class ColorFormatterMixin(BaseFormatter):
-    def _print_attrs(self, attrs, *values, **kwargs):
-        stream = kwargs.get("file", sys.stdout)
-        if stream.isatty() or self.force_color:
+    @property
+    def colored(self):
+        return sys.stdout.isatty() or self.force_color
+
+    def _print_attrs(self, attrs: str, *values: Any, **kwargs: Any) -> None:
+        file = kwargs.get("file", None)
+        if file is not None:
+            raise ValueError()
+        if self.colored:
             print(
                 "\033[0m\033[{attrs}m{value}\033[0m".format(
                     attrs=attrs, value="".join([str(value) for value in values])
@@ -163,24 +182,113 @@ class ColorFormatterMixin(BaseFormatter):
         else:
             print(*values, **kwargs)
 
-    def print_white(self, *values, **kwargs):
+    def print_white(self, *values: Any, **kwargs: Any) -> None:
         self._print_attrs("1", *values, **kwargs)
 
-    def print_green(self, *values, **kwargs):
+    def print_green(self, *values: Any, **kwargs: Any) -> None:
         self._print_attrs("32", *values, **kwargs)
 
-    def print_red(self, *values, **kwargs):
+    def print_red(self, *values: Any, **kwargs: Any) -> None:
         self._print_attrs("31", *values, **kwargs)
 
-    def print_yellow(self, *values, **kwargs):
+    def print_yellow(self, *values: Any, **kwargs: Any) -> None:
         self._print_attrs("33", *values, **kwargs)
 
-    def print_cyan(self, *values, **kwargs):
+    def print_cyan(self, *values: Any, **kwargs: Any) -> None:
         self._print_attrs("36", *values, **kwargs)
 
 
+class FailurePrinterMixin(ColorFormatterMixin):
+    def _get_test_module_index(self, tb: traceback.StackSummary) -> int:
+        test_module_index = len(tb) - 1
+
+        test_module_paths = [
+            import_module(import_module_name).__file__
+            for import_module_name in self.import_module_names
+        ]
+
+        for index, value in enumerate(tb):
+            path = value[0]
+            if path in test_module_paths:
+                if index < test_module_index:
+                    test_module_index = index
+
+        return test_module_index
+
+    def _print_stack_trace(self, exception: BaseException, cause_depth: int) -> None:
+        indent = "  " * cause_depth
+        if cause_depth:
+            self.print_red(f"{indent}    Caused by ", end="")
+
+        self.print_red(
+            "{exception_class}: {message}".format(
+                exception_class=exception.__class__.__name__,
+                message=f"\n{indent}    ".join(str(exception).split("\n")),
+            )
+        )
+
+        tb = traceback.extract_tb(exception.__traceback__)
+
+        test_module_index = self._get_test_module_index(tb)
+
+        for index, (path, line, function_name, text) in enumerate(tb):
+            if not self.show_testslide_stack_trace and index < test_module_index:
+                continue
+            if self.trim_path_prefix:
+                split = path.split(self.trim_path_prefix)
+                if len(split) == 2 and not split[0]:
+                    path = split[1]
+            row_text = (
+                '  File "{path}", line {line}, in {function_name}\n'
+                "    {text}\n".format(
+                    path=path,
+                    line=line,
+                    function_name=function_name,
+                    text=text,
+                )
+            )
+            if self.colored:
+                row_text = pygments.highlight(
+                    row_text,
+                    pygments.lexers.PythonTracebackLexer(),
+                    pygments.formatters.TerminalFormatter(),
+                )
+            row_text = "\n".join(
+                "{indent}    {line}".format(indent=indent, line=line)
+                for line in row_text.split("\n")[:-1]
+            )
+            print(row_text)
+
+        if exception.__cause__:
+            self._print_stack_trace(exception.__cause__, cause_depth=cause_depth + 1)
+
+    def print_failed_example(
+        self,
+        number: int,
+        example: Example,
+        exception: Union[Exception, AggregatedExceptions],
+    ) -> None:
+        self.print_white(
+            "  {number}) {context}: {example}".format(
+                number=number, context=example.context.full_name, example=example
+            )
+        )
+        if type(exception) is AggregatedExceptions:
+            exception_list = exception.exceptions  # type: ignore
+        else:
+            exception_list = [exception]
+        for number, exception in enumerate(exception_list):
+            self.print_red(
+                "    {number}) ".format(
+                    number=number + 1,
+                ),
+                end="",
+            )
+            self._print_stack_trace(exception, cause_depth=0)
+
+
 class SlowImportWarningMixin(ColorFormatterMixin):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         super().__init__(*args, **kwargs)
         if self.import_secs and self.import_secs > 1 and self._import_secs_warn:
             self.print_yellow(
@@ -193,11 +301,12 @@ class SlowImportWarningMixin(ColorFormatterMixin):
 
 
 class DSLDebugMixin:
-    def get_dsl_debug_indent(self, example):
+    def get_dsl_debug_indent(self, example: Example) -> str:
         return ""
 
-    def _dsl_print(self, example, description, code):
-        if not self.dsl_debug:
+    def _dsl_print(self, example: Example, description: str, code: Callable) -> None:
+        lineno: Union[str, int]
+        if not self.dsl_debug:  # type: ignore
             return
         name = code.__name__
         try:
@@ -207,17 +316,17 @@ class DSLDebugMixin:
                 file = inspect.getfile(code)
             except TypeError:
                 file = "?"
-        if file.startswith(os.path.dirname(__file__)):
+        if file and file.startswith(os.path.dirname(__file__)):
             return
-        if self.trim_path_prefix:
-            split = file.split(self.trim_path_prefix)
+        if self.trim_path_prefix:  # type: ignore
+            split = file.split(self.trim_path_prefix)  # type: ignore
             if len(split) == 2 and not split[0]:
                 file = split[1]
         try:
             _lines, lineno = inspect.getsourcelines(code)
         except OSError:
             lineno = "?"
-        self.print_cyan(
+        self.print_cyan(  # type: ignore
             "{indent}{description}: {name} @ {file_lineno}".format(
                 indent=self.get_dsl_debug_indent(example),
                 description=description,
@@ -226,25 +335,25 @@ class DSLDebugMixin:
             )
         )
 
-    def dsl_example(self, example, code):
+    def dsl_example(self, example: Example, code: Callable) -> None:
         self._dsl_print(example, "example", code)
 
-    def dsl_before(self, example, code):
+    def dsl_before(self, example: Example, code: Callable) -> None:
         self._dsl_print(example, "before", code)
 
-    def dsl_after(self, example, code):
+    def dsl_after(self, example: Example, code: Callable) -> None:
         self._dsl_print(example, "after", code)
 
-    def dsl_around(self, example, code):
+    def dsl_around(self, example: Example, code: Callable) -> None:
         self._dsl_print(example, "around", code)
 
-    def dsl_memoize(self, example, code):
+    def dsl_memoize(self, example: Example, code: Callable) -> None:
         self._dsl_print(example, "memoize", code)
 
-    def dsl_memoize_before(self, example, code):
+    def dsl_memoize_before(self, example: Example, code: Callable) -> None:
         self._dsl_print(example, "memoize_before", code)
 
-    def dsl_function(self, example, code):
+    def dsl_function(self, example: Example, code: Callable) -> None:
         self._dsl_print(example, "function", code)
 
 
@@ -257,46 +366,54 @@ class QuietFormatter(BaseFormatter):
     pass
 
 
-class ProgressFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
+class ProgressFormatter(DSLDebugMixin, SlowImportWarningMixin, FailurePrinterMixin):
     """
     Simple formatter that outputs "." when an example passes or "F" w
     """
 
-    def new_example(self, example):
+    def new_example(self, example: Example) -> None:
         super().new_example(example)
         if self.dsl_debug:
             print("")
 
-    def success(self, example):
+    def success(self, example: Example) -> None:
         super().success(example)
         self.print_green(".", end="")
 
-    def fail(self, example, exception):
+    def fail(self, example: Example, exception: BaseException) -> None:
         super().fail(example, exception)
         self.print_red("F", end="")
 
-    def skip(self, example):
+    def skip(self, example: Example) -> None:
         super().skip(example)
         self.print_yellow("S", end="")
 
-    def finish(self, not_executed_examples):
+    def finish(self, not_executed_examples: List[Example]) -> None:
         super().finish(not_executed_examples)
+        if self.results["fail"] and not self.dsl_debug:
+            self.print_red("\nFailures:")
+            for number, result in enumerate(self.results["fail"]):
+                result = cast(Dict[str, Union[Example, BaseException]], result)
+                print("")
+                self.print_failed_example(
+                    number + 1, result["example"], result["exception"]  # type: ignore
+                )
         print("")
 
 
-class DocumentFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
-    def get_dsl_debug_indent(self, example):
+class DocumentFormatter(DSLDebugMixin, SlowImportWarningMixin, FailurePrinterMixin):
+    def get_dsl_debug_indent(self, example: Example) -> str:
         return "  " * (example.context.depth + 1)
 
-    def new_context(self, context):
+    def new_context(self, context: Context) -> None:
         self.print_white(
             "{}{}{}".format("  " * context.depth, "*" if context.focus else "", context)
         )
 
-    def _color_output(self):
+    def _color_output(self) -> bool:
         return sys.stdout.isatty() or self.force_color
 
-    def success(self, example):
+    def success(self, example: Example) -> None:
         super().success(example)
         self.print_green(
             "{indent}{focus}{example}{pass_text}".format(
@@ -307,7 +424,7 @@ class DocumentFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMix
             )
         )
 
-    def fail(self, example, exception):
+    def fail(self, example: Example, exception: BaseException) -> None:
         if isinstance(exception, AggregatedExceptions) and 1 == len(
             exception.exceptions
         ):
@@ -325,7 +442,7 @@ class DocumentFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMix
             )
         )
 
-    def skip(self, example):
+    def skip(self, example: Example) -> None:
         super().skip(example)
         self.print_yellow(
             "{indent}{focus}{example}{skip_text}".format(
@@ -336,41 +453,7 @@ class DocumentFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMix
             )
         )
 
-    def print_failed_example(self, number, example, exception):
-        self.print_white(
-            "  {number}) {context}: {example}".format(
-                number=number, context=example.context.full_name, example=example
-            )
-        )
-        if type(exception) is AggregatedExceptions:
-            exception_list = exception.exceptions
-        else:
-            exception_list = [exception]
-        for number, exception in enumerate(exception_list):
-            self.print_red(
-                "    {number}) {exception_class}: {message}".format(
-                    number=number + 1,
-                    exception_class=exception.__class__.__name__,
-                    message="\n    ".join(str(exception).split("\n")),
-                )
-            )
-            for path, line, function_name, text in traceback.extract_tb(
-                exception.__traceback__
-            ):
-                if not self.show_testslide_stack_trace and path.startswith(
-                    os.path.dirname(__file__)
-                ):
-                    continue
-                if self.trim_path_prefix:
-                    split = path.split(self.trim_path_prefix)
-                    if len(split) == 2 and not split[0]:
-                        path = split[1]
-                self.print_cyan(
-                    '      File "{}", line {}, in {}\n'
-                    "        {}".format(path, line, function_name, text)
-                )
-
-    def finish(self, not_executed_examples):
+    def finish(self, not_executed_examples: List[Example]) -> None:
         super().finish(not_executed_examples)
         success = len(self.results["success"])
         fail = len(self.results["fail"])
@@ -379,16 +462,21 @@ class DocumentFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMix
         if self.results["fail"]:
             self.print_red("\nFailures:")
             for number, result in enumerate(self.results["fail"]):
+                result = cast(Dict[str, Union[Example, BaseException]], result)
                 print("")
-                self.print_failed_example(
-                    number + 1, result["example"], result["exception"]
+                self.print_failed_example(  # type: ignore
+                    number + 1, result["example"], result["exception"]  # type: ignore
                 )
         print("")
         self.print_white(
-            "Finished %s example(s) in %.1fs: ." % (total, self.duration_secs)
+            "Finished %s example(s) in %.1fs "
+            % (total, cast(float, self.duration_secs)),
+            end="",
         )
-        if self.import_secs > 2:
-            self.print_white("Imports took: %.1fs" % (self.import_secs))
+        if self.import_secs and self.import_secs > 2:
+            self.print_white("(Imports took: %.1fs)" % (self.import_secs))
+        else:
+            print("")
         if success:
             self.print_green("  Successful: ", success)
         if fail:
@@ -399,11 +487,11 @@ class DocumentFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMix
             self.print_cyan("  Not executed: ", len(not_executed_examples))
 
 
-class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
-    def get_dsl_debug_indent(self, example):
+class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, FailurePrinterMixin):
+    def get_dsl_debug_indent(self, example: Example) -> str:
         return "  "
 
-    def new_example(self, example):
+    def new_example(self, example: Example) -> None:
         self.print_white(
             "{}{}: ".format(
                 "*" if example.context.focus else "", example.context.full_name
@@ -413,10 +501,10 @@ class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
         if self.dsl_debug:
             print("")
 
-    def _color_output(self):
+    def _color_output(self) -> bool:
         return sys.stdout.isatty() or self.force_color
 
-    def success(self, example):
+    def success(self, example: Example) -> None:
         super().success(example)
         if self.dsl_debug:
             print("  ", end="")
@@ -428,7 +516,7 @@ class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
             )
         )
 
-    def fail(self, example, exception):
+    def fail(self, example: Example, exception: BaseException) -> None:
         if isinstance(exception, AggregatedExceptions) and 1 == len(
             exception.exceptions
         ):
@@ -439,7 +527,8 @@ class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
             print("  ", end="")
         self.print_red(
             "{focus}{example}: ".format(
-                focus="*" if example.focus else "", example=example,
+                focus="*" if example.focus else "",
+                example=example,
             ),
             end="",
         )
@@ -450,7 +539,7 @@ class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
             )
         )
 
-    def skip(self, example):
+    def skip(self, example: Example) -> None:
         super().skip(example)
         if self.dsl_debug:
             print("  ", end="")
@@ -462,41 +551,7 @@ class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
             )
         )
 
-    def print_failed_example(self, number, example, exception):
-        self.print_white(
-            "  {number}) {context}: {example}".format(
-                number=number, context=example.context.full_name, example=example
-            )
-        )
-        if type(exception) is AggregatedExceptions:
-            exception_list = exception.exceptions
-        else:
-            exception_list = [exception]
-        for number, exception in enumerate(exception_list):
-            self.print_red(
-                "    {number}) {exception_class}: {message}".format(
-                    number=number + 1,
-                    exception_class=exception.__class__.__name__,
-                    message="\n    ".join(str(exception).split("\n")),
-                )
-            )
-            for path, line, function_name, text in traceback.extract_tb(
-                exception.__traceback__
-            ):
-                if not self.show_testslide_stack_trace and path.startswith(
-                    os.path.dirname(__file__)
-                ):
-                    continue
-                if self.trim_path_prefix:
-                    split = path.split(self.trim_path_prefix)
-                    if len(split) == 2 and not split[0]:
-                        path = split[1]
-                self.print_cyan(
-                    '      File "{}", line {}, in {}\n'
-                    "        {}".format(path, line, function_name, text)
-                )
-
-    def finish(self, not_executed_examples):
+    def finish(self, not_executed_examples: List[Example]) -> None:
         super().finish(not_executed_examples)
         success = len(self.results["success"])
         fail = len(self.results["fail"])
@@ -505,16 +560,21 @@ class LongFormatter(DSLDebugMixin, SlowImportWarningMixin, ColorFormatterMixin):
         if self.results["fail"]:
             self.print_red("\nFailures:")
             for number, result in enumerate(self.results["fail"]):
+                result = cast(Dict[str, Union[Example, BaseException]], result)
                 print("")
-                self.print_failed_example(
-                    number + 1, result["example"], result["exception"]
+                self.print_failed_example(  # type: ignore
+                    number + 1, result["example"], result["exception"]  # type: ignore
                 )
         print("")
         self.print_white(
-            "Finished %s example(s) in %.1fs: ." % (total, self.duration_secs)
+            "Finished %s example(s) in %.1fs "
+            % (total, cast(float, self.duration_secs)),
+            end="",
         )
-        if self.import_secs > 2:
-            self.print_white("Imports took: %.1fs" % (self.import_secs))
+        if self.import_secs and self.import_secs > 2:
+            self.print_white("(Imports took: %.1fs)" % (self.import_secs))
+        else:
+            print("")
         if success:
             self.print_green("  Successful: ", success)
         if fail:
@@ -537,18 +597,18 @@ class Runner(object):
 
     def __init__(
         self,
-        contexts,
-        formatter,
-        shuffle=False,
-        seed=None,
-        focus=False,
-        fail_fast=False,
-        fail_if_focused=False,
-        names_text_filter=None,
-        names_regex_filter=None,
-        names_regex_exclude=None,
-        quiet=False,
-    ):
+        contexts: List[Context],
+        formatter: Union[SlowImportWarningMixin, DocumentFormatter],
+        shuffle: bool = False,
+        seed: int = None,
+        focus: bool = False,
+        fail_fast: bool = False,
+        fail_if_focused: bool = False,
+        names_text_filter: Optional[str] = None,
+        names_regex_filter: Optional[Pattern] = None,
+        names_regex_exclude: Optional[Pattern] = None,
+        quiet: bool = False,
+    ) -> None:
         self.contexts = contexts
         self.formatter = formatter
         self.shuffle = shuffle
@@ -561,7 +621,7 @@ class Runner(object):
         self.names_regex_exclude = names_regex_exclude
         self.quiet = quiet
 
-    def _run_example(self, example):
+    def _run_example(self, example: Example) -> None:
         if example.focus and self.fail_if_focused:
             raise AssertionError(
                 "Focused example not allowed with --fail-if-focused"
@@ -586,7 +646,7 @@ class Runner(object):
         else:
             _ExampleRunner(example, self.formatter).run()
 
-    def run(self):
+    def run(self) -> int:
         """
         Execute all examples in all contexts.
         """
@@ -620,7 +680,7 @@ class Runner(object):
         sys.stderr.flush()
         return exit_code
 
-    def _filter(self, example, focus):
+    def _filter(self, example: Example, focus: bool) -> bool:
         if focus and not example.focus:
             return False
 
@@ -639,7 +699,7 @@ class Runner(object):
         return True
 
     @property
-    def _all_examples(self):
+    def _all_examples(self) -> List[Example]:
         examples = [
             example for context in self.contexts for example in context.all_examples
         ]
@@ -650,7 +710,7 @@ class Runner(object):
         return examples
 
     @property
-    def _to_execute_examples(self):
+    def _to_execute_examples(self) -> List[Example]:
         examples = [
             example
             for example in self._all_examples
